/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#pragma once

#include <ox/model/fieldcounter.hpp>
#include <ox/model/optype.hpp>
#include <ox/model/typenamecatcher.hpp>
#include <ox/model/types.hpp>
#include <ox/std/buffer.hpp>
#include <ox/std/byteswap.hpp>
#include <ox/std/string.hpp>
#include <ox/std/trace.hpp>
#include <ox/std/vector.hpp>

#include "err.hpp"
#include "intops.hpp"
#include "presenceindicator.hpp"
#include "types.hpp"

namespace ox {

class MetalClawReader {

	private:
		FieldBitmapReader<const uint8_t*> m_fieldPresence;
		int m_fields = 0;
		int m_field = 0;
		int m_unionIdx = -1;
		std::size_t m_buffIt = 0;
		std::size_t m_buffLen = 0;
		const uint8_t *m_buff = nullptr;
		MetalClawReader *m_parent = nullptr;

	public:
		MetalClawReader(const uint8_t *buff, std::size_t buffLen, int unionIdx = -1, MetalClawReader *parent = nullptr) noexcept;

		~MetalClawReader() noexcept;

		Error field(const char*, int8_t *val) noexcept;
		Error field(const char*, int16_t *val) noexcept;
		Error field(const char*, int32_t *val) noexcept;
		Error field(const char*, int64_t *val) noexcept;

		Error field(const char*, uint8_t *val) noexcept;
		Error field(const char*, uint16_t *val) noexcept;
		Error field(const char*, uint32_t *val) noexcept;
		Error field(const char*, uint64_t *val) noexcept;

		Error field(const char*, bool *val) noexcept;

		// array handler
		template<typename T>
		Error field(const char*, T *val, std::size_t len) noexcept;

		// map handler
		template<typename T>
		Error field(const char*, HashMap<String, T> *val) noexcept;

		// array handler, with callback to allow handling individual elements
		template<typename T, typename Handler>
		Error field(const char*, Handler handler) noexcept;

		template<typename T>
		Error field(const char*, T *val) noexcept;

		template<typename U>
		Error field(const char*, UnionView<U> val) noexcept;

		template<std::size_t SmallStringSize>
		Error field(const char*, BasicString<SmallStringSize> *val) noexcept;

		template<std::size_t L>
		Error field(const char*, BString<L> *val) noexcept;

		Error field(const char*, SerStr val) noexcept;

		Error fieldCString(const char *name, char **val, int len) noexcept;

		/**
		 * Reads an array length from the current location in the buffer.
		 * @param pass indicates that the parsing should iterate past the array length
		 */
		Result<ArrayLength> arrayLength(const char *name, bool pass = true) noexcept;

		/**
		 * Reads an string length from the current location in the buffer.
		 */
		[[nodiscard]]
		StringLength stringLength(const char *name) noexcept;

		template<typename T = std::nullptr_t>
		constexpr void setTypeInfo() noexcept;

		constexpr void setTypeInfo(const char *name, int fields) noexcept;

		/**
		 * Returns a MetalClawReader to parse a child object.
		 */
		[[nodiscard]]
		MetalClawReader child(const char *name, int unionIdx = -1) noexcept;

		/**
		 * Indicates whether or not the next field to be read is present.
		 */
		[[nodiscard]]
		bool fieldPresent(const char *name) const noexcept;

		/**
		 * Indicates whether or not the given field is present.
		 */
		[[nodiscard]]
		bool fieldPresent(int fieldNo) const noexcept;

		void nextField() noexcept;

		[[nodiscard]]
		static constexpr auto opType() noexcept {
			return OpType::Read;
		}

	private:
		template<typename I>
		Error readInteger(I *val) noexcept;

};

template<typename T>
Error MetalClawReader::field(const char *name, T *val) noexcept {
	if constexpr(isVector_v<T>) {
		if (m_unionIdx == -1 || m_unionIdx == m_field) {
			// set size of val if the field is present, don't worry about it if not
			if (m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
				oxRequire(len, arrayLength(name, false));
				val->resize(len);
			}
			return field(name, val->data(), val->size());
		}
		++m_field;
		return OxError(0);
	} else {
		if ((m_unionIdx == -1 || m_unionIdx == m_field) && val && m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
			auto reader = child("");
			oxReturnError(model(&reader, val));
		}
		++m_field;
		return OxError(0);
	}
}

template<typename U>
Error MetalClawReader::field(const char*, UnionView<U> val) noexcept {
	if ((m_unionIdx == -1 || m_unionIdx == m_field) && val.get() && m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
		auto reader = child("", val.idx());
		oxReturnError(model(&reader, val.get()));
	}
	++m_field;
	return OxError(0);
}

template<std::size_t SmallStringSize>
Error MetalClawReader::field(const char*, BasicString<SmallStringSize> *val) noexcept {
	if (m_unionIdx == -1 || m_unionIdx == m_field) {
		if (m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
			// read the length
			if (m_buffIt >= m_buffLen) {
				return OxError(MC_BUFFENDED);
			}
			std::size_t bytesRead = 0;
			oxRequire(size, mc::decodeInteger<StringLength>(&m_buff[m_buffIt], m_buffLen - m_buffIt, &bytesRead));
			m_buffIt += bytesRead;
			const auto cap = size;
			*val = BasicString<SmallStringSize>(cap);
			auto data = val->data();
			// read the string
			if (static_cast<StringLength>(cap) < size) {
				return OxError(MC_OUTBUFFENDED);
			}
			if (m_buffIt + size > m_buffLen) {
				return OxError(MC_BUFFENDED);
			}
			memcpy(data, &m_buff[m_buffIt], size);
			data[size] = 0;
			m_buffIt += size;
		} else {
			*val = "";
		}
	}
	++m_field;
	return OxError(0);
}

template<std::size_t L>
Error MetalClawReader::field(const char *name, BString<L> *val) noexcept {
	return field(name, SerStr(val->data(), val->cap()));
}

template<typename I>
Error MetalClawReader::readInteger(I *val) noexcept {
	if (m_unionIdx == -1 || m_unionIdx == m_field) {
		if (m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
			std::size_t bytesRead = 0;
			if (m_buffIt >= m_buffLen) {
				oxTrace("ox::MetalClaw::readInteger") << "Buffer ended";
				return OxError(MC_BUFFENDED);
			}
			auto valErr = mc::decodeInteger<I>(&m_buff[m_buffIt], m_buffLen - m_buffIt, &bytesRead);
			m_buffIt += bytesRead;
			oxReturnError(valErr.error);
			*val = valErr.value;
		} else {
			*val = 0;
		}
	}
	++m_field;
	return OxError(0);
}

// array handler
template<typename T>
Error MetalClawReader::field(const char *name, T *val, std::size_t valLen) noexcept {
	if (m_unionIdx == -1 || m_unionIdx == m_field) {
		if (m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
			// read the length
			if (m_buffIt >= m_buffLen) {
				return OxError(MC_BUFFENDED);
			}
			std::size_t bytesRead = 0;
			oxRequire(len, mc::decodeInteger<ArrayLength>(&m_buff[m_buffIt], m_buffLen - m_buffIt, &bytesRead));
			m_buffIt += bytesRead;

			// read the list
			if (valLen >= len) {
				auto reader = child("");
				reader.setTypeInfo("List", static_cast<int>(len));
				for (std::size_t i = 0; i < len; ++i) {
					oxReturnError(reader.field("", &val[i]));
				}
			} else {
				oxTrace("ox::mc::read::field(T)") << name << ", size:" << valLen;
				return OxError(MC_OUTBUFFENDED);
			}
		}
	}
	++m_field;
	return OxError(0);
}

template<typename T>
Error MetalClawReader::field(const char*, HashMap<String, T> *val) noexcept {
	if (m_unionIdx == -1 || m_unionIdx == m_field) {
		if (m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
			// read the length
			if (m_buffIt >= m_buffLen) {
				return OxError(MC_BUFFENDED);
			}
			std::size_t bytesRead = 0;
			oxRequire(len, mc::decodeInteger<ArrayLength>(&m_buff[m_buffIt], m_buffLen - m_buffIt, &bytesRead));
			m_buffIt += bytesRead;
			// read the list
			auto reader = child("");
			reader.setTypeInfo("List", static_cast<int>(len));
			for (std::size_t i = 0; i < len; ++i) {
				const auto keyLen = reader.stringLength(nullptr);
				auto wkey = ox_malloca(keyLen + 1, char, 0);
				auto wkeyPtr = wkey.get();
				oxReturnError(reader.fieldCString("", &wkeyPtr, static_cast<int>(keyLen + 1)));
				oxReturnError(reader.field("", &val->operator[](wkey.get())));
			}
		}
	}
	++m_field;
	return OxError(0);
}

template<typename T, typename Handler>
Error MetalClawReader::field(const char*, Handler handler) noexcept {
	if (m_unionIdx == -1 || m_unionIdx == m_field) {
		if (m_fieldPresence.get(static_cast<std::size_t>(m_field))) {
			// read the length
			if (m_buffIt >= m_buffLen) {
				return OxError(MC_BUFFENDED);
			}
			std::size_t bytesRead = 0;
			oxRequire(len, mc::decodeInteger<ArrayLength>(&m_buff[m_buffIt], m_buffLen - m_buffIt, &bytesRead));
			m_buffIt += bytesRead;

			// read the list
			auto reader = child("");
			reader.setTypeInfo("List", static_cast<int>(len));
			for (std::size_t i = 0; i < len; ++i) {
				T val;
				oxReturnError(reader.field("", &val));
				oxReturnError(handler(i, &val));
			}
		}
	}
	++m_field;
	return OxError(0);
}

template<typename T>
constexpr void MetalClawReader::setTypeInfo() noexcept {
	setTypeInfo(ModelTypeName_v<T>, countFields<T>());
}

constexpr void MetalClawReader::setTypeInfo(const char*, int fields) noexcept {
	m_fields = fields;
	m_buffIt = static_cast<std::size_t>((fields / 8 + 1) - (fields % 8 == 0));
	m_fieldPresence.setFields(fields);
	m_fieldPresence.setMaxLen(static_cast<int>(m_buffIt));
}

template<typename T>
Error readMC(const char *buff, std::size_t buffLen, T *val) noexcept {
	MetalClawReader reader(reinterpret_cast<const uint8_t*>(buff), buffLen);
	return model(&reader, val);
}

template<typename T>
Result<T> readMC(const char *buff, std::size_t buffLen) noexcept {
	T val;
	oxReturnError(readMC(buff, buffLen, &val));
	return val;
}

template<typename T>
Result<T> readMC(const Buffer &buff) noexcept {
	return readMC<T>(buff.data(), buff.size());
}

}
