/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <filesystem>

#include <ox/std/std.hpp>

#include "project.hpp"

namespace nostalgia::studio {

ox::String filePathToName(const ox::String &path, const ox::String &prefix, const ox::String &suffix) noexcept {
	const auto begin = prefix.len();
	const auto end = path.len() - (suffix.len() + prefix.len());
	return path.substr(begin, end);
}


Project::Project(ox::FileSystem *fs, const ox::String &path) noexcept: m_fs(fs) {
	oxTracef("nostalgia::studio", "Project: {}", path);
	m_path = path;
	buildFileIndex();
}

ox::Error Project::create() noexcept {
	std::error_code ec;
	std::filesystem::create_directory(m_path.toStdString(), ec);
	return OxError(ec.value(), "PassThroughFS: mkdir failed");
}

ox::FileSystem *Project::romFs() noexcept {
	return m_fs;
}

ox::Error Project::mkdir(const ox::String &path) const noexcept {
	oxReturnError(m_fs->mkdir(path.c_str(), true));
	fileUpdated.emit(path);
	return OxError(0);
}

ox::Result<ox::FileStat> Project::stat(const ox::String &path) const noexcept {
	return m_fs->stat(path.c_str());
}

bool Project::exists(const ox::String &path) const noexcept {
	return m_fs->stat(path.c_str()).error == 0;
}

const ox::Vector<ox::String> &Project::fileList(const char *ng) noexcept {
	return m_fileExtFileMap[ng];
}

void Project::buildFileIndex() noexcept {
	auto [files, err] = listFiles();
	if (err) {
		oxLogError(err);
		return;
	}
	std::sort(files.begin(), files.end());
	for (const auto &file : files) {
		if (!file.beginsWith("/.nostalgia/")) {
			const auto [ext, err] = fileExt(file);
			if (err) {
				continue;
			}
			m_fileExtFileMap[ext].emplace_back(file);
		}
	}
}

ox::Error Project::writeBuff(const ox::String &path, const ox::Buffer &buff) const noexcept {
	oxReturnError(m_fs->write(path.c_str(), buff.data(), buff.size()));
	fileUpdated.emit(path);
	return OxError(0);
}

ox::Result<ox::Buffer> Project::loadBuff(const ox::String &path) const noexcept {
	return m_fs->read(path.c_str());
}

ox::Error Project::lsProcDir(ox::Vector<ox::String> *paths, const ox::String &path) const noexcept {
	oxRequire(files, m_fs->ls(path.c_str()));
	for (const auto &name : files) {
		const auto fullPath = path + "/" + name.c_str();
		oxRequire(stat, m_fs->stat(fullPath.c_str()));
		switch (stat.fileType) {
			case ox::FileType::NormalFile:
				paths->push_back(fullPath);
				break;
			case ox::FileType::Directory:
				oxReturnError(lsProcDir(paths, fullPath));
				break;
			case ox::FileType::None:
				break;
		}
	}
	return OxError(0);
}

ox::Result<ox::Vector<ox::String>> Project::listFiles(const ox::String &path) const noexcept {
	ox::Vector<ox::String> paths;
	oxReturnError(lsProcDir(&paths, path));
	return paths;
}

}
